import sys
import CreateQuantDataFrames
import pandas
import msparser
import io
import WriteReports
import LoadQuantitation

# create an msparser logger
mylogger_ = msparser.ms_stdout_logger()

def main():

    if len(sys.argv) < 2 :
        sys.exit('Must specify properties filename as parameters')

    propsPath = sys.argv[1]
    # Load in properties.csv - this is the settings file written out to Python from Distiller
    # with the report configuration to run with
    props_csv = pandas.read_csv(propsPath, delimiter = ',', header = 0, keep_default_na = False, quotechar="\"")

    # Create subsets of the properties file
    props_logging = props_csv[props_csv.Identifier == 'LoggingOptions'] # Log and column mask
    props_options = props_csv[props_csv.Identifier == 'PathOptions'] # paths to mascot/cgi, msparser, reports output
    
    props_header = props_csv[props_csv.Identifier == 'ReportHeader'] # writing report
    props_rawfiles = props_csv[props_csv.Identifier == 'RawFile'] # writing report
    props_selectionType = props_csv[props_csv.Identifier == 'selectionType'] # Peptide selection criteria for the top-3 calculation.  See top-3.py.xml
    peptideSelectionType = props_selectionType.Input1.iloc[0]
    props_exportHeader = props_csv[props_csv.Identifier == 'exportHeader'] # export a header at the top of the report.  See top-3.py.xml
    exportHeader = props_exportHeader.Input1.iloc[0]
    props_analysis = props_csv[props_csv.Identifier == 'AnalysisInputs'] # standard inputs for the report
    excludeDatabase = float(props_analysis.Input5.values[0]) if props_analysis.Input5.values[0] != '' else '' #(numeric) Database to exclude in analysis.  If 0, all databases are used
    savePath = props_options.Input2.iloc[0]    # file path to save the report to.

    # The following properties are not used in this script, but are listed here to show how to access them
    props_actions = props_csv[props_csv.Identifier == 'ActionOptions'] # do create matrix, do create graphs, do create html
    props_results = props_csv[props_csv.Identifier == 'Results'] # .dat, .xml, .cdb, cache
    props_schemas = props_csv[props_csv.Identifier == 'Schema'] # The Distiller XML schema path
    props_multidats = props_csv[props_csv.Identifier == 'MultiDats'] # 1.dat, 2.dat, 3.dat, ... for multifile projects
    props_summary = props_csv[props_csv.Identifier == 'SummaryInputs'] # creating pepSum flag settings
    props_instruments = props_csv[props_csv.Identifier == 'Instrument'] # writing report
    props_export = props_csv[props_csv.Identifier == 'ExportFileType'] # file type to write, used if the report has a choice of output formats

    # Initialise logging.  See msparser API help for details
    mylogger_.setColsToOutput(int(props_logging.Input2.iloc[0]))
    msparser.ms_loggingmonitor.getDefaultMonitor().setLogMask(int(props_logging.Input1.iloc[0]))
    msparser.ms_loggingmonitor.getDefaultMonitor().addLogEventsHandler(mylogger_)

    # load the peptide summary, quantitation results etc using the helper method in the supplied LoadQuantitation file
    loadRes = LoadQuantitation.DoLoad(propsPath)

    # call the function to generate the report
    CreateTop3Report(loadRes, excludeDatabase, peptideSelectionType, savePath, props_header, props_rawfiles, exportHeader)

    # make a call to the logger to prevent garbage collection. See "Garbage collection problems (advanced reading)"
    # in the Mascot Parser documentation.
    mylogger_.setColsToOutput(1)

## Creates the top-3 intensity report
# \param loadedResults The loaded search and quantitation results.
# \param excludeDatabase The contaminants database index to exclude from the report (if set)
# \param peptideSelectionType The peptide selection/grouping criteria selected by the user on the report Wizard
# \param savePath The report export path defined by Distiller
# \param props_header Header properties passed in from Distiller
# \param props_rawfiles Raw file properties information passed in from Distiller
# \param exportHeader String set to "True" or "False" depending on the user selection on the report Wizard
def CreateTop3Report(loadedResults, excludeDatabase, peptideSelectionType, savePath, props_header, props_rawfiles, exportHeader):
    # Check we have results
    if (len(loadedResults) == 0):
        sys.exit("No Results Parameters Supplied")

    # Get the msparser ms_peptidesummary, quantitation results back from the loadedResults array
    isMS1 = loadedResults[0].isMS1 # MS1 quantitation results?  True or false
    quant = loadedResults[0].qObj # The quantitation results (either ms_ms1quantitation or ms_ms2quantitation depending on whether or not isMS1 is true)
    pepSum = loadedResults[0].pepSum # The ms_peptidesummary (protein hits)
    qMethod = loadedResults[0].qMethod # The quantitation method definition (ms_quant_method)

    # MS2, Precursor and Average quantitation methods all require different handling to find the individual peptide component intensities
    isAverage = False
    if isMS1:
        isAverage = qMethod.getProtocol().getAverage() != None    

    if not isMS1:
        # disable normalisation for MS2 quantitation
        quant.getQuantitationMethod().dropNormalisation()
    
    # Get the component names
    componentNames = []
    if isAverage:
        componentNames.append('Avg')
    else:
        for c in range(0,qMethod.getNumberOfComponents()):
            component = qMethod.getComponentByNumber(c).getName()
            componentNames.append(component)
    
    # Helper method which pulls the anchor protein ms_protein results from the peptide summary into an array.
    proteins = CreateQuantDataFrames.pullProteinsFrom(pepSum)

    # if the data are from an average quant calculation, we can do most of the required steps directly
    # through msparser, so update the quant method to ensure we match the report settings
    if isAverage:
        qMethod.getProtocol().getAverage().setSelection(peptideSelectionType)
        qMethod.getProtocol().getAverage().setNumPeptides(3)

    end = len(proteins)
    # this will be an array of arrays containing the data to export to the CSV file
    data = []
    for i in range(0, end):
        # output progress information
        WriteReports.OutputProgress('Calculating protein intensity values', (i+1), end)

        row = []
        protein = proteins[i]

        # skip any protein hits from the selected contaminants database
        if protein.getDB() == excludeDatabase:
            continue
        
        if not isMS1 and not isAverage:
            # load in the first protein ratio to initialise the ms_ms2quantitation results for this protein
            for r in range(0, 1):#quant.getQuantitationMethod().getNumberOfReportRatios()):
                ratioName = quant.getQuantitationMethod().getReportRatioByNumber(r).getName()
                proteinRatio = quant.getProteinRatio(protein.getAccession(), protein.getDB(), ratioName)
                ratMessage = '{} {}'.format(ratioName, proteinRatio.getValue())
                msparser.ms_loggingmonitor.getDefaultMonitor().logMessage(msparser.ms_loggingmonitor.LVL_DEBUG3, msparser.ms_loggingmonitor.SRC_PURIFIERAPP, 0, ratMessage,'top-3.py', sys._getframe().f_lineno)

        hitNo = protein.getHitNumber()
        if protein.getMemberNumber() > 0:
            row.append("{}.{}".format(hitNo,protein.getMemberNumber()))
        else:
            row.append(hitNo)
        row.append(protein.getAccession())
                
        for name in enumerate(componentNames):
            intensity = calculateProteinIntensity(protein, pepSum, quant, name, isAverage, isMS1, peptideSelectionType)
            if intensity > 0:
                row.append(intensity)
            else:
                row.append('')

        row.append(pepSum.getProteinDescription(protein.getAccession()))
        data.append(row)

    # create a pandas DataFrame from our data array with column headers
    header = []
    header.append('')
    header.append('Accession')    
    for name in componentNames:
        header.append("{} {}".format(name,'intensity'))
    header.append('Description')
        
    # pull the data into a Pandas DataFrame
    # to easily output a CSV file with a header but no row index
    df = pandas.DataFrame(data, None, header)
    if exportHeader == 'True':
        # output the report to an intermediate string
        s = io.StringIO()
        # generate a report header
        csvRow = '{0},{1}\n'
        # for examples of other available header information, look at the createHeader method
        # in the supplied reportHeader.py script
        s.write(csvRow.format('Project file', props_header.Input2.iloc[0]))
        numFiles = props_rawfiles.shape[0]
        datfiles = props_header.Input3.iloc[0].split(",")
        for i in range(0, numFiles):
            s.write(csvRow.format('Raw file '+str(i+1), props_rawfiles.Input1.iloc[0]))
            if len(datfiles) == numFiles:
                s.write(csvRow.format('Mascot search result '+str(i+1), datfiles[i]))
        s.write('\n')
        # now write the main report table        
        df.to_csv(s, index=False)
        pagepd = open(savePath, 'w', encoding='utf-8')
        pagepd.write(s.getvalue())
        pagepd.close()
    else:
        # if we don't want to add a header, can write directly to the file path rather than a string:
        df.to_csv(savePath, ',','',None,None,True,False)

## Top level method to calculate the top-3 average intensity for the specified protein and component
# \param protein The ms_protein to calculate the average intensity for
# \param peptideSummary The msparser ms_peptidesummary results
# \param quantResults The msparser quantitation results (ms_ms1quantitation or ms_ms2quantitation depending on the quantitation method)
# \param componentName The component to calculate the intensity for
# \param isAverage boolean - True if the quantResults are the Average [MD] method
# \param isMS1 boolean - True is the quantResults are ms_ms1quantitation results
# \param peptideSelection The peptide selection/grouping criteria selected by the user on the report Wizard
# \return The calculated intensity, or 0 if the intensity cannot be calculated
def calculateProteinIntensity(protein, peptideSummary, quantResults, componentName, isAverage, isMS1, peptideSelectionType):
    if isAverage:
        return calculateProteinIntensityAverage(protein, quantResults)
    elif isMS1:
        return calculateProteinIntensityMS1(protein, quantResults, componentName, peptideSelectionType)
    else:
        return calculateProteinIntensityMS2(protein, peptideSummary, quantResults, componentName, peptideSelectionType)

## Uses msparser to calculate the top-3 average protein intensity for the passed ms_protein
# \param protein The ms_protein to calculate the average intensity for
# \param quantResults The msparser quantitation results
# \return The calculated intensity, or 0 if the intensity cannot be calculated
def calculateProteinIntensityAverage(protein, quantResults):
    # we can use msparser to calculate the values
    include = msparser.ms_peptide_quant_key_vector()
    exclude = msparser.ms_peptide_quant_key_vector()
    sampleSize = 0
    # getAveragePeptideIntensity only works if the quantitation method is Average
    # In Python this results an array, the first value is True or False depending on whether or not
    # the calculation was successful (sufficient peptides matching the criteria)
    # the second value in the intensity
    intensity = quantResults.getAveragePeptideIntensity(protein.getAccession(), protein.getDB(), include, exclude)
    if intensity[0] == True:
        return intensity[1]
    else:
        return 0

## Calculates the top-3 average protein intensity for the specied protein and component for Reporter and Multiplex methods
# \param protein The ms_protein to calculate the average intensity for
# \param peptideSummary The msparser ms_peptidesummary results
# \param quantResults The msparser quantitation results (ms_ms1quantitation or ms_ms2quantitation depending on the quantitation method)
# \param componentName The component to calculate the intensity for
# \param peptideSelection The peptide selection/grouping criteria selected by the user on the report Wizard
# \return The calculated intensity, or 0 if the intensity cannot be calculated
def calculateProteinIntensityMS2(protein, peptideSummary, quantResults, componentName, peptideSelectionType):
    # get the peptide keys which pass the quality thresholds
    peptideKeys = quantResults.getPeptideQuantKeys(protein.getAccession(), protein.getDB())    
    # find the index of the component
    componentIndex = componentName[0];    
    groupIntensities = {}
    for i in range(0, peptideKeys.size()):
        intensities = quantResults.getComponentIntensities(peptideKeys.get(i))        
        key = ''
        ok, query, rank = peptideKeys.get(i).getQueryAndRank()
        peptide = peptideSummary.getPeptide(query, rank)
        if peptideSelectionType == 'unique_sequence':
            key = peptide.getPeptideStr()
        elif peptideSelectionType == 'unique_mr':
            key = peptide.getPeptideStr()+'_'+peptideSummary.getReadableVarMods(query, rank)
        elif peptideSelectionType == 'unique_mz':
            key = '{}_{}_{}'.format(peptide.getPeptideStr(), peptideSummary.getReadableVarMods(query, rank), peptide.getCharge())
        
        if key in groupIntensities:
            groupIntensities[key] = groupIntensities[key] + intensities[componentIndex]
        else:
            groupIntensities[key] = intensities[componentIndex]

    if len(groupIntensities.keys()) >= 3:
        # we have enough keys, so calculate the average
        intensity = 0
        sortedIntensities = sorted(groupIntensities.values(), reverse=True)
        for i in range(0, 3):
            intensity += sortedIntensities[i]
        return intensity / 3
    
    return 0

## Calculates the top-3 average protein intensity for the specied protein and component for Precursor methods
# \param protein The ms_protein to calculate the average intensity for
# \param quantResults The msparser quantitation results (ms_ms1quantitation or ms_ms2quantitation depending on the quantitation method)
# \param componentName The component to calculate the intensity for
# \param peptideSelection The peptide selection/grouping criteria selected by the user on the report Wizard
# \return The calculated intensity, or 0 if the intensity cannot be calculated
def calculateProteinIntensityMS1(protein, quantResults, componentName, peptideSelectionType):
    # get the protein quantitation peptide matches
    matchesCount = quantResults.getNumMatchesForHit(protein.getHitNumber(), protein.getMemberNumber())

    qm = quantResults.getQuantitationMethod()    
    quality = qm.getQuality()
    matchedFractionThreshold = 1
    if quality.isIsolatedPrecursor():
        matchedFractionThreshold = float(quality.getIsolatedPrecursorThreshold())
    totalIntensityThreshold = 1
    if quality.isTotalIntensity():
        totalIntensityThreshold = float(quality.getTotalIntensityThreshold())
    correlationThreshold = float(qm.getIntegration().getMatchedRho())

    groupIntensities = {}
    for i in range(1, matchesCount+1):
        match = quantResults.getMatchForHit(i, protein.getHitNumber(), protein.getMemberNumber())
        
        key = ''
        if peptideSelectionType == 'unique_sequence':
            key = match.getPeptideString()
        elif peptideSelectionType == 'unique_mr':
            key = match.getPeptideString()+'_'+match.getReadableLabelFreeVarMods()
        elif peptideSelectionType == 'unique_mz':
            key = '{}_{}_{}'.format(match.getPeptideString(), match.getReadableLabelFreeVarMods(), match.getChargeState())
        component = match.getComponent(componentName[1])
        # check for various quality thresholds which apply to the entire match rather than
        # individual ratios
        if quality.isIsolatedPrecursor() and match.getChargeStateData().getMatchedFraction() < matchedFractionThreshold:
            continue
        if quality.isTotalIntensity() and match.getChargeStateData().getTotalIntensity() < totalIntensityThreshold:
            continue
        if match.getChargeStateData().getMatchedRho() < correlationThreshold:
            continue

        # note, this is the unnormalised value
        if key in groupIntensities:
            groupIntensities[key] = groupIntensities[key] + component.getAbsoluteValue()
        else:
            groupIntensities[key] = component.getAbsoluteValue()

    if len(groupIntensities.keys()) >= 3:
        # we have enough keys, so calculate the average
        intensity = 0
        sortedIntensities = sorted(groupIntensities.values(), reverse=True)
        for i in range(0, 3):
            intensity += sortedIntensities[i]
        return intensity / 3
    
    return 0


if (__name__ == "__main__"):
    sys.exit(main())